;; Generated package description from org-jami-bot.el  -*- no-byte-compile: t -*-
(define-package "org-jami-bot" "0.0.5" "Capture GNU Jami messages as notes and todos in Org mode" '((emacs "28.1") (jami-bot "0.0.4")) :commit "020b03f299dad438f65d7bcbf93553b273fd7c33" :authors '(("Hanno Perrey" . "hanno@hoowl.se")) :maintainer '("Hanno Perrey" . "hanno@hoowl.se") :keywords '("comm" "outlines" "org-capture" "jami") :url "https://gitlab.com/hperrey/org-jami-bot")
